/*
 * Copyright 2016 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "globus_gridftp_server_s3_stat.h"

static
S3Status
globus_l_s3_stat_head_dir_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg);

static
void
globus_l_s3_stat_head_dir_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data);

static
S3Status
globus_l_s3_stat_list_dir_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg);

static
void
globus_l_s3_stat_list_dir_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data);

static
S3Status
globus_l_s3_stat_list_dir_callback(
    int                                 is_truncated,
    const char                         *next_marker,
    int                                 contents_count, 
    const S3ListBucketContent          *contents,
    int                                 common_prefixes_count,
    const char                        **common_prefixes,
    void                               *callback_data);

static
S3Status
globus_l_s3_stat_head_object_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg);

static
void
globus_l_s3_stat_head_object_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data);

void
globus_i_s3_stat_object(
    globus_i_s3_stat_t                 *stat_state)
{
    char                                keydir[strlen(stat_state->key)+2];

    GlobusS3Enter();

    sprintf(keydir, "%s/", stat_state->key);

    stat_state->retries = 0;
    do
    {
        stat_state->status = S3StatusOK;
        stat_state->result = GLOBUS_SUCCESS;

        globus_i_s3_retry_delay(stat_state->retries);

        /* Head "key/" to see if we created a directory placeholder */
        GlobusS3Info("REST: op=%p S3_head_object(bucket=%s, key=%s) try=%d\n",
                    (void *) stat_state->op,
                    stat_state->bucket_name,
                    keydir,
                    stat_state->retries);
        S3_head_object(
                &stat_state->bucket_context,
                keydir,
                NULL,
                &(S3ResponseHandler)
                {
                    .propertiesCallback   = globus_l_s3_stat_head_dir_properties,
                    .completeCallback     = globus_l_s3_stat_head_dir_complete
                },
                stat_state);

        GlobusS3DebugOpStatus(S3_head_object, stat_state);
    }
    while (S3_status_is_retryable(stat_state->status)
        && (stat_state->retries++) < globus_i_s3_max_retries);

    if (stat_state->result != GLOBUS_SUCCESS
        || stat_state->stat_count > 0)
    {
        return;
    }

    stat_state->retries = 0;
    do
    {
        stat_state->status = S3StatusOK;
        stat_state->result = GLOBUS_SUCCESS;

        globus_i_s3_retry_delay(stat_state->retries);

        /* List "key/" to see if it's the prefix to a key or other prefixes*/
        GlobusS3Info("REST: op=%p S3_list_bucket(bucket=%s, key=%s) try=%d\n",
                    (void *) stat_state->op,
                    stat_state->bucket_name,
                    keydir,
                    stat_state->retries);
        S3_list_bucket(
                &stat_state->bucket_context,
                keydir,
                NULL,
                "/",
                1,
                NULL,
                &(S3ListBucketHandler)
                {
                    .responseHandler = (S3ResponseHandler)
                    {
                        .propertiesCallback
                                      = globus_l_s3_stat_list_dir_properties,
                        .completeCallback
                                      = globus_l_s3_stat_list_dir_complete
                    },
                    .listBucketCallback
                                      = globus_l_s3_stat_list_dir_callback
                },
                stat_state);

        GlobusS3DebugOpStatus(S3_list_bucket, stat_state);
    }
    while (S3_status_is_retryable(stat_state->status)
        && (stat_state->retries++) < globus_i_s3_max_retries);

    if (stat_state->result != GLOBUS_SUCCESS
        || stat_state->stat_count > 0)
    {
        return;
    }

    /* Head key to see if it's a file */
    stat_state->retries = 0;

    do
    {
        stat_state->status = S3StatusOK;
        stat_state->result = GLOBUS_SUCCESS;

        globus_i_s3_retry_delay(stat_state->retries);

        GlobusS3Info("REST: op=%p S3_head_object(bucket=%s, key=%s) try=%d\n",
                    (void *) stat_state->op,
                    stat_state->bucket_name,
                    stat_state->key,
                    stat_state->retries);

        S3_head_object(
                &stat_state->bucket_context,
                stat_state->key,
                NULL,
                &(S3ResponseHandler)
                {
                    .propertiesCallback
                                      = globus_l_s3_stat_head_object_properties,
                    .completeCallback = globus_l_s3_stat_head_object_complete
                },
                stat_state);

        GlobusS3DebugOpStatus(S3_head_object, stat_state);
    }
    while (S3_status_is_retryable(stat_state->status)
        && (stat_state->retries++) < globus_i_s3_max_retries);

    GlobusS3Exit();
}
/* globus_l_s3_stat_object() */

static
S3Status
globus_l_s3_stat_head_dir_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg)
{
    S3Status                            status = S3StatusOK;
    globus_i_s3_stat_t                 *stat_state = callback_arg;


    GlobusS3Enter();


    if (properties != NULL
        && properties->contentType != NULL
        && strcmp(properties->contentType, GLOBUS_S3_DIR_PLACEHOLDER_TYPE) == 0)
    {
        char                                dirname[strlen(stat_state->stat_info->pathname) + 2];
        int                                 dev, ino;

        snprintf(
                dirname,
                sizeof(dirname),
                "%s",
                stat_state->stat_info->pathname);

        dev = globus_hashtable_string_hash(stat_state->bucket_name, INT_MAX);
        ino = globus_hashtable_string_hash(dirname, INT_MAX);

        globus_i_s3_stat_add(
                stat_state,
                &(globus_gfs_stat_t)
                {
                    .mode             = S_IRWXU|S_IFDIR,
                    .nlink            = 1,
                    .name             = ".",
                    .dev              = dev,
                    .ino              = ino,
                    .mtime            = (time_t) -1,
                    .atime            = (time_t) -1,
                    .ctime            = (time_t) -1
                });
    }


    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_stat_head_dir_properties() */

static
void
globus_l_s3_stat_head_dir_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data)
{
    globus_i_s3_stat_t                 *stat_state = callback_data;

    GlobusS3Enter();

    if (status != S3StatusOK && status != S3StatusHttpErrorNotFound)
    {
        stat_state->status = status;
        stat_state->result = 
                GlobusGFSErrorS3(
                    status,
                    stat_state->stat_info->pathname,
                    error_details);
    }
    GlobusS3Exit();
}
/* globus_l_s3_stat_head_dir_complete() */

static
S3Status
globus_l_s3_stat_list_dir_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg)
{
    S3Status                            status = S3StatusOK;

    GlobusS3Enter();

    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_stat_list_dir_properties() */

static
void
globus_l_s3_stat_list_dir_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data)
{
    globus_i_s3_stat_t                 *stat_state = callback_data;

    GlobusS3Enter();

    if (status != S3StatusOK && status != S3StatusHttpErrorNotFound && stat_state->result == GLOBUS_SUCCESS)
    {
        stat_state->status = status;
        stat_state->result = 
                GlobusGFSErrorS3(
                    status,
                    stat_state->stat_info->pathname,
                    error_details);
    }

    GlobusS3Exit();
}
/* globus_l_s3_stat_list_dir_complete() */

static
S3Status
globus_l_s3_stat_list_dir_callback(
    int                                 is_truncated,
    const char                         *next_marker,
    int                                 contents_count, 
    const S3ListBucketContent          *contents,
    int                                 common_prefixes_count,
    const char                        **common_prefixes,
    void                               *callback_data)
{

    globus_i_s3_stat_t                 *stat_state = callback_data;
    S3Status                            status = S3StatusOK;
    char                                dirname[strlen(stat_state->key) + 2];
    int                                 dev, ino;


    GlobusS3Enter();

    snprintf(dirname, sizeof(dirname), "%s/", stat_state->key);

    if ((contents_count > 0
            && strncmp(contents[0].key, dirname,strlen(dirname)) == 0)
        || common_prefixes_count > 0)
    {
        dev = globus_hashtable_string_hash(
                (void *) stat_state->bucket_name,
                INT_MAX);
        ino = globus_hashtable_string_hash(
                (void *) dirname,
                INT_MAX);

        globus_i_s3_stat_add(
                stat_state,
                &(globus_gfs_stat_t)
                {
                    .mode             = S_IRWXU|S_IFDIR,
                    .nlink            = 1,
                    .name             = ".",
                    .dev              = dev,
                    .ino              = ino,
                    .mtime            = (time_t) -1,
                    .atime            = (time_t) -1,
                    .ctime            = (time_t) -1
                });

    }
    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_stat_list_dir_callback() */

static
S3Status
globus_l_s3_stat_head_object_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg)
{
    GlobusS3Enter();

    globus_i_s3_stat_t                 *stat_state = callback_arg;
    S3Status                            status = S3StatusOK;

    if (properties != NULL)
    {
        int                                 dev, ino;

        dev = globus_hashtable_string_hash(stat_state->bucket_name, INT_MAX);
        ino = globus_hashtable_string_hash(stat_state->stat_info->pathname, INT_MAX);

        globus_i_s3_stat_add(
                stat_state,
                &(globus_gfs_stat_t)
                {
                    .mode             = S_IRWXU|S_IFREG,
                    .nlink            = 1,
                    .name             = stat_state->key,
                    .dev              = dev,
                    .ino              = ino,
                    .size             = (globus_off_t) ((properties->contentLength < 0) ? (int64_t) -1 : (int64_t) properties->contentLength),
                    .mtime            = (time_t) (properties->lastModified < 0 ? -1 : properties->lastModified),
                    .atime            = (time_t) -1,
                    .ctime            = (time_t) -1
                });
    }

    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_stat_head_object_properties() */

static
void
globus_l_s3_stat_head_object_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data)
{
    globus_i_s3_stat_t                 *stat_state = callback_data;

    GlobusS3Enter();

    if (status != S3StatusOK && stat_state->result == GLOBUS_SUCCESS)
    {
        stat_state->status = status;
        stat_state->result = GlobusGFSErrorS3(status, stat_state->stat_info->pathname, error_details);
    }
    GlobusS3Exit();
}
/* globus_l_s3_stat_head_object_complete() */

// vim: ft=c:
